<?php

namespace Baqend\WordPress\Controller;

use Baqend\SDK\Model\AssetFilter;
use Baqend\WordPress\Admin\MenuBuilder;
use Baqend\WordPress\Admin\View;
use Baqend\WordPress\Loader;
use Baqend\WordPress\Plugin;
use Psr\Log\LoggerInterface;

/**
 * Class AdminController created on 16.06.17.
 *
 * @author Konstantin Simon Maria Möllers
 * @package Baqend\WordPress\Controller
 */
class AdminController extends Controller {

    /**
     * @var View
     */
    private $view;

    /**
     * @var string
     */
    private $settings_group;

    /**
     * Admin constructor.
     *
     * @param Plugin $plugin
     */
    public function __construct( Plugin $plugin, LoggerInterface $logger ) {
        parent::__construct( $plugin, $logger );
        $this->view           = new View();
        $this->settings_group = $plugin->slug . '_group';
    }

    public function register( Loader $loader ) {
        $loader->add_action( 'admin_enqueue_scripts', [ $this, 'assets' ] );
        $loader->add_action( 'admin_init', [ $this, 'register_settings' ] );
        $loader->add_action( 'admin_menu', [ $this, 'add_menus' ] );
        $loader->add_action( 'admin_notices', [ $this, 'add_notices' ] );
        $loader->add_filter( 'plugin_action_links_' . $this->plugin->get_basename(), [ $this, 'add_settings_link' ] );
    }

    public function assets() {
        $version = $this->plugin->version;

        wp_enqueue_style( $this->plugin->slug . '-iqons', $this->plugin->icons_url( 'baqend.css' ), [], $version );
        wp_enqueue_style( $this->plugin->slug, $this->plugin->css_url( 'baqend-admin.css' ), [ 'wp-codemirror' ], $version );
        wp_enqueue_script( 'baqend-vendor', $this->plugin->js_url( 'baqend-vendor.js' ), [ 'jquery' ], $version, true );
        wp_enqueue_script( 'baqend-admin', $this->plugin->js_url( 'baqend-admin.js' ), [ 'jquery', 'wp-codemirror', ], $version, true );
    }

    public function register_settings() {
        if ( ! $this->plugin->app_name && $this->is_on_baqend_admin( [ 'baqend', 'baqend_help' ] ) ) {
            exit( wp_redirect( baqend_admin_url( 'baqend' ) ) );
        }

        register_setting( $this->settings_group, $this->plugin->slug, [ $this, 'save_settings' ] );
    }

    /**
     * Adds links to the plugin page.
     *
     * @param string[] $links
     *
     * @return string[]
     */
    public function add_settings_link( array $links ) {
        $url           = $this->plugin->is_plesk_user() ? baqend_admin_url( 'baqend_speed_kit' ) : baqend_admin_url( 'baqend' );
        $settings_link = '<a href="' . $url . '">' . __( 'Settings', 'baqend' ) . '</a>';
        array_push( $links, $settings_link );

        return $links;
    }

    /**
     * Processes options aver saving them.
     *
     * FIXME: This is also called when saving the options. What should we do?
     *
     * @param string[] $options Options to process.
     *
     * @return mixed[] The processed options.
     */
    public function save_settings( array $options = null ) {
        $fields  = $this->create_fields();
        $changed = [];

        foreach ( $fields as $field ) {
            $key = $field['slug'];
            if ( ! array_key_exists( $key, $options ) ) {
                continue;
            }

            $oldValue = $this->plugin->options->get( $key );
            $value    = $options[ $key ];

            if ( is_string( $value ) ) {
                $value = $this->normalize_setting( $field, $value );
            }

            if ( $oldValue !== $value ) {
                $this->plugin->options->set( $key, $value );
                $changed[ $key ] = $value;
            }

            // clears cache of fastest cache
            if ( isset( $GLOBALS['wp_fastest_cache'] ) &&
                 method_exists( $GLOBALS['wp_fastest_cache'], 'deleteCache' ) ) {
                $GLOBALS['wp_fastest_cache']->deleteCache();
            }

            // clears cache of simple cache
            if ( has_action( 'sc_purge_cache' ) ) {
                do_action( 'sc_purge_cache' );
            }
        }

        $this->update_cron_hooks( $changed );

        return $this->plugin->options->all();
    }

    /**
     * Register the administration menu for this plugin into the WordPress Dashboard menu.
     *
     * @return array
     */
    public function add_menus() {
        $stats = $this->stats();

        $menu = new MenuBuilder();

        $menu
            ->title( __( 'Speed Kit', 'baqend' ) )
            ->slug( 'baqend' )
            ->icon( 'dashicons-speed-kit' );

        if ( ! $this->plugin->is_plesk_user() ) {
            // Add settings submenu item
            $menu->entry( __( 'Overview', 'baqend' ), 'baqend', [ $this, 'render_overview' ] );
        }

        // Add deploy submenu item
        $menu->entry( __( 'Settings', 'baqend' ), 'baqend_speed_kit', [ $this, 'render_speed_kit' ] );

        $menu->entry( __( 'Advanced', 'baqend' ), 'baqend_advanced', [ $this, 'render_advanced' ] );

        if ( $this->plugin->is_plesk_user() ) {
            $menu->main( 'baqend_speed_kit' );
        } else {
            // Add account submenu item
            $menu->entry( __( 'Account', 'baqend' ), 'baqend_account', [ $this, 'render_account' ] );

            $menu->main( 'baqend' );
        }

        // Add help submenu item
        $menu->entry( __( 'Help', 'baqend' ), 'baqend_help', [ $this, 'render_help' ] );

        if ( ! $this->plugin->is_plesk_user() && $stats && ( $stats->is_free() || $stats->is_exceeded() ) ) {
            // Add upgrade submenu item
            $menu->entry( __( 'Upgrade', 'baqend' ), 'baqend_upgrade', [ $this, 'render_upgrade' ], true );
        }

        $this->view->set_tabs( $menu->build() );
    }

    /**
     * Displays notices in the admin if applicable.
     */
    public function add_notices() {
        if ( $this->is_login_info_broken() ) {
            ?>
            <div class="notice notice-error is-dismissible">
                <p>
                    <strong><?php _e( 'We updated the Baqend WordPress plugin. Please logout and login again to fix a login vulnerability.', 'baqend' ) ?></strong>
                </p>
                <p>
                    <a href="<?php echo baqend_admin_url( 'baqend_account' ); ?>"><?php _e( 'Click here to go to the logout', 'baqend' ) ?></a>
                </p>
            </div>
            <?php
        }

        if ( ! is_ssl() && strpos( site_url(), 'http://localhost' ) === false ) {
            ?>
            <div class="notice notice-warning is-dismissible">
                <p>
                    <strong><?php _e( 'Baqend Speed Kit needs HTTPS to work correctly.', 'baqend' ) ?></strong>
                </p>
                <p>
                    <?php _e( 'Please ensure your WordPress is running on HTTPS.', 'baqend' ) ?>
                    <?php _e( 'Therefore, install an SSL certificate on your Apache or Nginx web server, or ask your website provider to install it for you.', 'baqend' ) ?>
                    <?php _e( 'Then, go to the general settings and ensure your WordPress address is beginning with “https://”.', 'baqend' ) ?>
                </p>
                <p>
                    <a href="<?php echo admin_url( 'options-general.php' ); ?>"><?php _e( 'Click here to go to the general settings', 'baqend' ) ?></a>
                </p>
            </div>
            <?php
        }

        if ( $this->plugin->options->has( 'cron_error' ) ) {
            $cron_error = $this->plugin->options->get( 'cron_error' );
            ?>
            <div class="notice notice-error is-dismissible" data-dismiss-ajax="clear_cron_error">
                <p><strong><?php echo $cron_error; ?></strong></p>
            </div>
            <?php
        }

        $stats             = $this->stats();
        $is_exceeded       = $stats ? $stats->is_exceeded() : false;
        $speed_kit_enabled = $this->plugin->options->get( 'speed_kit_enabled' );
        if ( $is_exceeded && $speed_kit_enabled ) {
            ?>
            <div class="notice notice-warning is-dismissible">
                <p>
                    <strong><?php _e( 'Your free plan is exceeded. Speed Kit is no longer active.', 'baqend' ) ?></strong>
                </p>
                <p>
                    <?php _e( 'Your current plan has been exceeded.', 'baqend' ) ?>
                    <?php _e( 'Therefore, Speed Kit is not activated for your website.', 'baqend' ) ?>
                    <?php _e( 'Upgrade your plan to gain the performance boost of Speed Kit for your currently hosted WordPress blog.', 'baqend' ) ?>
                </p>
                <p>
                    <a href="<?php echo baqend_admin_url( 'baqend_upgrade' ); ?>"><?php _e( 'Click here to upgrade your plan', 'baqend' ) ?></a>
                </p>
            </div>
            <?php
        }
    }

    /**
     * Render the overview view.
     */
    public function render_overview() {
        $stats             = $this->stats();
        $is_exceeded       = $stats ? $stats->is_exceeded() : false;
        $speed_kit_enabled = $this->plugin->options->get( 'speed_kit_enabled' );
        $comparison        = $this->plugin->analyzer_service->load_latest_comparison( $speed_kit_enabled, $is_exceeded );

        if ( $comparison === null ) {
            $this->plugin->analyzer_service->start_comparison( $speed_kit_enabled );
        }

//        $stats = new Stats(['appName' => '', 'fixPrice' => 0, 'requestsNow' => 114e3, 'requestsMax' => 150e3, 'priceMax' => 10000, 'priceNow' => 400, 'trafficMax' => 4*1024*1024*1024, 'trafficNow' => 4*1024*1024*1024]);

        $this->view
            ->set_template( 'overview.php' )
            ->assign( 'logged_in', $this->plugin->app_name !== null )
            ->assign( 'app_name', $this->plugin->app_name )
            ->assign( 'username', $this->plugin->options->get( 'username' ) )
            ->assign( 'speed_kit', $speed_kit_enabled )
            ->assign( 'exceeded', $is_exceeded )
            ->assign( 'stats', $stats )
            ->assign( 'comparison', $comparison )
            ->assign( 'bbq_username', $this->plugin->options->get( 'bbq_username' ) )
            ->assign( 'bbq_password', $this->plugin->options->get( 'bbq_password' ) )
            ->render();
    }

    /**
     * Render the account view.
     */
    public function render_account() {
        $settings  = $this->plugin->options->all();
        $api_token = $this->plugin->options->get( 'api_token' );

        $this->view
            ->set_template( 'account.php' )
            ->assign( 'app_name', $this->plugin->app_name )
            ->assign( 'stats', $this->stats() )
            ->assign( 'api_token', $api_token )
            ->assign( 'settings', $settings )
            ->assign( 'settings_group', $this->settings_group )
            ->assign( 'bbq_username', $this->plugin->options->get( 'bbq_username' ) )
            ->assign( 'bbq_password', $this->plugin->options->get( 'bbq_password' ) )
            ->render();
    }

    /**
     * Render the Speed Kit view.
     */
    public function render_speed_kit() {
        $template = 'speedKit.php';
        $slugs    = [
            'enabled_pages',
            'enabled_paths',
            'speed_kit_whitelist',
            'speed_kit_blacklist',
            'image_optimization',
            'speed_kit_cookies',
            'speed_kit_content_type',
            'speed_kit_update_interval',
            'fetch_origin_interval',
        ];

        if ( ! $this->plugin->is_plesk_user() ) {
            $slugs[] = 'speed_kit_enabled';
        }

        $this->render_settings( $slugs, $template );
    }

    public function render_advanced() {
        $template = 'advanced.php';
        $slugs    = [
            'speed_kit_max_staleness',
            'speed_kit_app_domain',
            'dynamic_block_config',
        ];

        $this->render_settings( $slugs, $template );
    }

    /**
     * Render the help view.
     */
    public function render_help() {
        // Generate the help fields
        $this->view
            ->set_template( 'help.php' )
            ->assign( 'logged_in', $this->plugin->app_name !== null )
            ->assign( 'stats', $this->stats() )
            ->render();
    }

    /**
     * Render the help view.
     */
    public function render_upgrade() {
        // Generate the upgrade fields
        $this->view
            ->set_template( 'upgrade.php' )
            ->assign( 'app_name', $this->plugin->app_name )
            ->assign( 'stats', $this->stats() )
            ->render();
    }

    private function render_settings( array $slugs, $template ) {
        // Speed Kit settings have changed, perform revalidation and cache invalidation
        if ( isset( $_GET['settings-updated'] ) && $_GET['settings-updated'] === 'true' ) {
            try {
                $filter = new AssetFilter();
                $filter->addContentType( AssetFilter::DOCUMENT );

                $this->plugin->baqend->asset()->revalidate( $filter );

                // Trigger a cache update for WP total cache and co
                $this->plugin->trigger_wp_cache_invalidation();

                $this->logger->info( 'Revalidation by save settings succeeded', [ 'filter' => $filter->jsonSerialize() ] );
            } catch ( \Exception $e ) {
                $this->logger->error( 'Revalidation by save settings failed with ' . get_class( $e ) . ': ' . $e->getMessage(), [ 'exception' => $e ] );
            }
        }

        $settings = $this->plugin->options->all();

        // Get updates on Speed Kit
        $sw_path      = $this->plugin->sw_path();
        $snippet_path = $this->plugin->snippet_path();
        $df_path      = $this->plugin->dynamic_fetcher_path();
        $info         = $this->plugin->baqend->getSpeedKit()->createInfo( $sw_path, $snippet_path, $df_path, '1.X.X' );

        $field_args = $this->create_fields( $slugs );
        $fields     = $this->plugin->form_service->render_fields( $field_args, $settings );

        $this->view
            ->set_template( $template )
            ->assign( 'stats', $this->stats() )
            ->assign( 'fields', $fields )
            ->assign( 'settings', $settings )
            ->assign( 'settings_group', $this->settings_group )
            ->assign( 'speed_kit', $info )
            ->render();
    }

    /**
     * Checks whether the user is on one of the given Baqend admin pages.
     *
     * @param string[] $pages The pages which should be checked.
     *
     * @return bool True, if user is on one of those pages.
     */
    private function is_on_baqend_admin( array $pages ) {
        global $pagenow;
        if ( $pagenow !== 'admin.php' || ! isset( $_GET['page'] ) ) {
            return false;
        }

        $page = $_GET['page'];

        if ( strpos( $page, 'baqend' ) !== 0 ) {
            return false;
        }

        return ! in_array( $page, $pages, true );
    }

    /**
     * @param string[]|null $fields_to_filter Slugs of the fields to display.
     *
     * @return array The fields.
     */
    private function create_fields( array $fields_to_filter = null ) {
        $fields = include __DIR__ . '/../../config/fields.php';
        if ( null === $fields_to_filter ) {
            return $fields;
        }

        // Filters only the field to display
        return array_filter( $fields, function ( $field ) use ( $fields_to_filter ) {
            return in_array( $field['slug'], $fields_to_filter, true );
        } );
    }

    public function update_cron_hooks( array $changed ) {
        $time = time();
        if ( isset( $changed['speed_kit_update_interval'] ) ) {
            wp_clear_scheduled_hook( 'cron_revalidate_html' );

            $speed_kit_update_interval = $changed['speed_kit_update_interval'];
            if ( $speed_kit_update_interval !== 'off' ) {
                wp_schedule_event( $time, $speed_kit_update_interval, 'cron_revalidate_html' );
            }
        }
    }

    /**
     * Normalizes a setting based on its field information.
     *
     * @param array $field The field information.
     * @param string $value The value to normalize.
     *
     * @return mixed
     */
    private function normalize_setting( array $field, $value ) {
        $type = $field['type'];

        if ( $type == 'list' ) {
            $value = preg_split( '/\r\n|[\r\n]/', $value );

            return array_filter( $value, function ( $item ) {
                return ! empty( $item );
            } );
        }

        if ( $type == 'checkbox' ) {
            return $value === 'true' ? true : ( $value === 'false' ? false : null );
        }

        return $value;
    }

    /**
     * Checks whether the login info is broken.
     *
     * @return bool
     */
    private function is_login_info_broken() {
        $options = $this->plugin->options;

        if ( ! $options->has( 'app_name' ) ) {
            return false;
        }

        if ( $options->has( 'api_token' ) ) {
            return false;
        }

        return ! $options->has( 'username' );
    }

    private function stats() {
        return $this->plugin->stats_service->load_stats();
    }
}
