<?php

namespace Baqend\WordPress\Controller;

use Baqend\SDK\Model\AssetFilter;
use Baqend\WordPress\Loader;

/**
 * Class Frontend created on 17.07.17.
 *
 * @author Konstantin Simon Maria Möllers
 * @package Baqend\WordPress\Controller
 */
class FrontendController extends Controller {

    public function register( Loader $loader ) {
        $loader->add_action( 'wp_head', [ $this, 'head' ], 1 );
        $loader->add_action( 'admin_head', [ $this, 'head' ], 1 );
        $loader->add_action( 'send_headers', [ $this, 'send_headers' ] );
    }

    /**
     * Is executed when WordPress is enqueueing scripts for the website.
     */
    public function head() {
        $app_name = $this->plugin->options->get( 'app_name' );
        if ( empty( $app_name ) ) {
            return;
        }

        // Find default whitelisted URLs
        $whitelisted_urls = [ home_url(), site_url( 'wp-content' ), site_url( 'wp-includes/js' ) ];
        $whitelisted_urls = strip_protocol( $whitelisted_urls );

        // Find default blacklisted URLs
        $blacklisted_urls = [
            site_url( 'wp-admin' ),
            site_url( 'wp-json' ),
            site_url( 'wp-login' ),
            site_url( 'login' ),
            site_url( 'wp-content/plugins/baqend' ),
        ];
        $blacklisted_urls = strip_protocol( $blacklisted_urls );

        $sw_url = $this->get_service_worker_url();
        $config = [
            'appName'   => $app_name,
            'sw'        => $sw_url,
            'scope'     => $this->plugin->sw_scope(),
            'disabled'  => is_user_logged_in() || ! $this->plugin->options->get( 'speed_kit_enabled' ),
            'whitelist' => [ [ 'url' => $whitelisted_urls ] ],
            'blacklist' => [ [ 'url' => $blacklisted_urls ] ],
        ];

        $enabled_sites_option = $this->get_enabled_sites_option();
        if ( ! empty( $enabled_sites_option ) ) {
            $config['enabledSites'] = $enabled_sites_option;
        }

        $image_optimization_option = $this->plugin->options->get( 'image_optimization' );
        if ( ! empty( $image_optimization_option ) && ! isset( $image_optimization_option['disabled'] ) || ! $image_optimization_option['disabled'] ) {
            $config['image'] = [
                'quality' => (int) $image_optimization_option['quality'],
                'webp'    => ! ! $image_optimization_option['webp'],
                'pjpeg'   => ! ! $image_optimization_option['pjpeg'],
            ];
        } else {
            $config['image'] = [
                'quality' => false,
                'webp'    => false,
                'pjpeg'   => false,
            ];
        }

        $whitelist_option = $this->plugin->options->get( 'speed_kit_whitelist' );
        if ( ! empty( $whitelist_option ) ) {
            $config['whitelist'][0]['url'] = array_merge(
                $config['whitelist'][0]['url'],
                $this->strip_comments( $whitelist_option )
            );
        }

        $config['whitelist'][0]['url'] = $this->simplify_rules( $config['whitelist'][0]['url'] );

        $content_types_option = $this->plugin->options->get( 'speed_kit_content_type' );
        if ( ! empty( $content_types_option ) ) {
            $config['whitelist'][0]['contentType'] = $content_types_option;
        }

        $blacklist_option = $this->plugin->options->get( 'speed_kit_blacklist' );
        if ( ! empty( $blacklist_option ) ) {
            $config['blacklist'][0]['url'] = array_merge(
                $config['blacklist'][0]['url'],
                $this->strip_comments( $blacklist_option )
            );
        }

        $config['blacklist'][0]['url'] = $this->simplify_rules( $config['blacklist'][0]['url'] );

        $woo_commerce_blacklist = $this->get_woo_commerce_blacklist();
        if ( ! empty( $woo_commerce_blacklist ) ) {
            if ( ! isset( $config['blacklist'] ) ) {
                $config['blacklist'] = [];
            }

            $config['blacklist'][] = $woo_commerce_blacklist;
        }

        $cookies_option = $this->plugin->options->get( 'speed_kit_cookies' );
        if ( ! empty( $cookies_option ) ) {
            if ( ! isset( $config['blacklist'] ) ) {
                $config['blacklist'] = [];
            }

            $config['blacklist'][] = [
                'cookie'      => $this->strip_comments( $cookies_option ),
                'contentType' => [ AssetFilter::DOCUMENT ],
            ];
        }

        $max_staleness_option = $this->plugin->options->get( 'speed_kit_max_staleness' );
        if ( 60000 !== (int) $max_staleness_option ) {
            $config['maxStaleness'] = (int) $max_staleness_option;
        }

        $app_domain_option = $this->plugin->options->get( 'speed_kit_app_domain' );
        if ( ! empty( $app_domain_option ) ) {
            $config['appDomain'] = $app_domain_option;
        }

        $fetch_origin_interval_option = (int) $this->plugin->options->get( 'fetch_origin_interval' );
        if ( - 1 !== $fetch_origin_interval_option ) {
            // Retrieve setting from PHP settings
            if ( - 2 === $fetch_origin_interval_option ) {
                $session_max_lifetime = ini_get( 'session.gc_maxlifetime' );
                if ( is_numeric( $session_max_lifetime ) ) {
                    $fetch_origin_interval_option = (int) $session_max_lifetime;
                } else {
                    $fetch_origin_interval_option = 1200;
                }
            }

            $config['fetchOriginInterval'] = $fetch_origin_interval_option;
        }

        $config  = json_encode( $config );
        $snippet = file_get_contents( $this->plugin->snippet_path() );
        echo <<<HTML
<!-- Baqend Speed Kit -->
<script type="application/javascript">(function(){var speedKit=$config;$snippet})();</script>

HTML;

        $dynamic_block_config_option = trim( $this->plugin->options->get( 'dynamic_block_config' ) );
        if ( ! empty( $dynamic_block_config_option ) ) {
            $dynamic_fetcher = file_get_contents( $this->plugin->dynamic_fetcher_path() );
            echo <<<HTML
<!-- Baqend Dynamic Fetcher -->
<script type="application/javascript">(function(){{$dynamic_block_config_option}
;{$dynamic_fetcher}})();</script>

HTML;
        }
    }

    /**
     * Is executed when WordPress wants to set HTTP response headers.
     */
    public function send_headers() {
        if ( headers_sent() || ! $this->plugin->baqend->isConnected() ) {
            return;
        }

        $sw_url   = $this->get_service_worker_url();
        $sw_scope = $this->plugin->sw_scope();
        header( 'Link: <' . $sw_url . '>; rel=preload; as=serviceworker; crossorigin', false );
        header( 'Link: <' . $sw_url . '>; rel=serviceworker; scope="' . $sw_scope . '"', false );
    }

    /**
     * Generates the "enabledSites" option for the Speed Kit config.
     *
     * @return array
     */
    private function get_enabled_sites_option() {
        $pathnames = [];

        $enabled_paths = $this->plugin->options->get( 'enabled_paths' );
        if ( is_array( $enabled_paths ) && ! empty( $enabled_paths ) ) {
            $pathnames = array_merge( $pathnames, $this->strip_comments( $enabled_paths ) );
        }

        $enabled_pages = $this->plugin->options->get( 'enabled_pages' );
        if ( is_array( $enabled_pages ) && ! empty( $enabled_pages ) ) {
            foreach ( $enabled_pages as $page ) {
                switch ( $page['type'] ) {
                    case 'Page':
                        $pathnames[] = $this->to_pathname_rule( get_permalink( $page['id'] ) );
                        break;
                    case 'Categories':
                        $posts       = get_posts( [ 'category' => $page['id'] ] );
                        $pathnames[] = $this->to_pathname_rule( get_category_link( $page['id'] ) );
                        foreach ( $posts as $post ) {
                            $pathnames[] = $this->to_pathname_rule( get_permalink( $post ) );
                        }
                        break;
                }
            }
        }

        if ( empty( $pathnames ) ) {
            return [];
        }

        return [ [ 'pathname' => $pathnames ] ];
    }

    /**
     * @return string
     */
    private function get_service_worker_url() {
        return $this->plugin->sw_url();
    }

    /**
     * @return array
     */
    private function get_woo_commerce_blacklist() {
        if ( ! $this->plugin->woo_commerce_service->is_shop_active() ) {
            return [];
        }

        $rules = $this->plugin->woo_commerce_service->load_woo_commerce_rules();

        return [
            'url'         => $this->wrap( reg_exp_escape( $rules ), 'regexp:/^', '$/' ),
            'contentType' => [ 'document' ],
        ];
    }

    /**
     * @param string[] $options
     *
     * @return string[]
     */
    private function strip_comments( array $options ) {
        return array_values( array_filter(
            array_map(
                function ( $option ) {
                    $index = strpos( $option, ';' );
                    if ( $index !== false ) {
                        $option = substr( $option, 0, $index );
                    }

                    return trim( $option );
                },
                $options
            ),
            function ( $option ) {
                return ! empty( $option );
            }
        ) );
    }

    private function reg_exp_escape( $subject ) {
        return preg_replace( '/[[\]{}()*+?.,\\^$|#\s\/]/', '\\\\$0', $subject );
    }

    /**
     * Wrap a string or a list of strings with a prefix and a suffix.
     *
     * @param string|string[] $subject
     * @param string $prefix
     * @param string $suffix
     *
     * @return string|string[]
     */
    private function wrap( $subject, $prefix, $suffix = '' ) {
        if ( is_array( $subject ) ) {
            return array_map(
                function ( $text ) use ( $prefix, $suffix ) {
                    return $this->wrap( $text, $prefix, $suffix );
                },
                $subject
            );
        }

        return $prefix . $subject . $suffix;
    }

    /**
     * @param string[] $rules
     *
     * @return string[]
     */
    private function simplify_rules( array $rules ) {
        $keep = [];
        foreach ( $rules as $rule ) {
            foreach ( $keep as $kept_rule ) {
                if ( strpos( $rule, $kept_rule ) === 0 ) {
                    continue 2;
                }
            }

            foreach ( $keep as $index => $kept_rule ) {
                if ( strpos( $kept_rule, $rule ) === 0 ) {
                    unset( $keep[ $index ] );
                }
            }

            $keep[] = $rule;
        }

        return array_values( $keep );
    }

    private function to_pathname_rule( $url ) {
        return 'regexp:/^' . reg_exp_escape( $this->strip_origin( $url ) ) . '$/';
    }

    private function strip_origin( $url ) {
        return str_replace( home_url(), '', $url );
    }
}
