<?php

namespace Baqend\WordPress\Controller;

use Baqend\SDK\Model\AssetFilter;
use Baqend\WordPress\Loader;
use WP_Post as Post;

/**
 * Class TriggerController created on 19.07.17.
 *
 * @author Konstantin Simon Maria Möllers
 * @package Baqend\WordPress\Controller
 */
class TriggerController extends Controller {

    const COMMENT_APPROVED = 1;
    const COMMENT_DECLINED = 0;

    public function register( Loader $loader ) {
        // For posts
        $loader->add_action( 'save_post', [ $this, 'revalidate_post' ] );
        $loader->add_action( 'delete_post', [ $this, 'revalidate_post' ] );
        $loader->add_action( 'add_attachment', [ $this, 'revalidate_post' ] );
        $loader->add_action( 'attachment_updated', [ $this, 'revalidate_post' ] );
        $loader->add_action( 'delete_attachment', [ $this, 'revalidate_post' ] );
        $loader->add_action( 'set_object_terms', [ $this, 'revalidate_post' ] );

        // For comments
        $loader->add_action( 'comment_post', [ $this, 'revalidate_comment' ] );
        $loader->add_action( 'edit_comment', [ $this, 'revalidate_comment' ] );
        $loader->add_action( 'delete_comment', [ $this, 'revalidate_comment' ] );
        $loader->add_action( 'transition_comment_status', [ $this, 'transition_comment' ], 10, 3 );

        // For terms
        $loader->add_action( 'edited_term', [ $this, 'revalidate_term' ] );
        $loader->add_action( 'delete_term', [ $this, 'revalidate_term' ] );

        // For users
        $loader->add_action( 'profile_update', [ $this, 'revalidate_user' ] );
        $loader->add_action( 'user_register', [ $this, 'revalidate_user' ] );
        $loader->add_action( 'delete_user', [ $this, 'revalidate_user' ] );

        // For themes
        $loader->add_action( 'switch_theme', [ $this, 'revalidate_site' ] );

        // For widgets
        $loader->add_filter( 'widget_update_callback', [ $this, 'widget_update_callback' ], 10, 2 );
        $loader->add_action( 'delete_widget', [ $this, 'revalidate_site' ] );

        // For updates
        $loader->add_action( 'upgrader_process_complete', [ $this, 'upgrader_process_complete' ] );
    }

    /**
     * @param array $instance
     * @param array $new_instance
     *
     * @return array
     */
    public function widget_update_callback( $instance, $new_instance ) {
        $this->revalidate_site();

        return $new_instance;
    }

    /**
     * Used by hooks to revalidate a post.
     *
     * @param int $post_ID
     */
    public function revalidate_post( $post_ID ) {
        $post = get_post( (int) $post_ID );
        if ( ! $post ) {
            return;
        }

        // For updated posts, send a revalidation request to the server
        $this->do_revalidate_post( $post );
    }

    /**
     * Used by hooks to revalidate a comment.
     *
     * @param int $comment_ID
     */
    public function revalidate_comment( $comment_ID ) {
        $comment = get_comment( $comment_ID );
        if ( ! $comment ) {
            return;
        }

        $post_id = $comment->comment_post_ID;
        $this->revalidate_post( $post_id );
    }

    /**
     * @param int|string $new_status
     * @param int|string $old_status
     * @param object $comment
     */
    public function transition_comment( $new_status, $old_status, $comment ) {
        $post_id = $comment->{'comment_post_ID'};
        $post    = get_post( $post_id );
        if ( ! $post ) {
            return;
        }

        $this->do_revalidate_post( $post );
    }

    /**
     * Revalidates a given term by its ID.
     *
     * @param int $term_ID
     */
    public function revalidate_term( $term_ID ) {
        $term = get_term( (int) $term_ID );
        if ( ! $term ) {
            return;
        }

        $filter = new AssetFilter();
        // Invalidate the home URL
        $filter->addUrl( home_url() );
        // Invalidate the term's permalink
        $filter->addUrl( get_term_link( $term ) );
        // Invalidate the feed
        $filter->addUrl( get_feed_link() );
        // Invalidate all archives
        $archive_URLs = $this->get_all_archive_urls();
        if ( ! empty( $archive_URLs ) ) {
            $filter->addUrls( $archive_URLs );
        }

        $this->send_revalidation( $filter );
    }

    /**
     * Revalidates a given user by its ID.
     *
     * @param int $user_ID
     */
    public function revalidate_user( $user_ID ) {
        $user = get_author_posts_url( (int) $user_ID );
        if ( ! $user ) {
            return;
        }

        $filter = new AssetFilter();
        // Invalidate the home URL
        $filter->addUrl( home_url() );
        // Invalidate the user's permalink
        $filter->addUrl( get_author_posts_url( $user_ID ) );
        // Invalidate the feed
        $filter->addUrl( get_feed_link() );
        // Invalidate all archives
        $archive_URLs = $this->get_all_archive_urls();
        if ( ! empty( $archive_URLs ) ) {
            $filter->addUrls( $archive_URLs );
        }

        $this->send_revalidation( $filter );
    }

    /**
     * Called when WordPress updated this plugin.
     */
    public function upgrader_process_complete() {
        try {
            // Check all files are downloaded after plugin update
            $this->plugin->speed_kit_service->ensure_files_up_to_date();

            if ( ! $this->plugin->baqend->isConnected() ) {
                return;
            }

            // Let other caching plugins revalidate the generated HTML
            $this->plugin->trigger_wp_cache_invalidation();

            // Revalidate HTML
            $filter = new AssetFilter();

            $filter->addContentTypes( [ AssetFilter::DOCUMENT ] );

            $this->plugin->baqend->asset()->revalidate( $filter );
            $this->logger->debug( 'Revalidation by plugin update succeeded', [ 'filter' => $filter->jsonSerialize() ] );
        } catch ( \Exception $e ) {
            $this->logger->error( 'Revalidation by plugin update failed with ' . get_class( $e ) . ': ' . $e->getMessage(), [ 'exception' => $e ] );
        }
    }

    /**
     * Revalidates the whole WordPress blog.
     */
    public function revalidate_site() {
        $site_url = site_url();

        $filter = new AssetFilter();
        $filter->addPrefix( $site_url );

        $this->send_revalidation( $filter );
    }

    /**
     * @param Post $post
     */
    private function do_revalidate_post( Post $post ) {
        $filter = new AssetFilter();
        // Invalidate the home URL
        $filter->addUrl( home_url() );
        // Invalidate the post's permalink
        $filter->addUrl( get_permalink( $post ) );
        // Invalidate the feed
        $filter->addUrl( get_feed_link() );
        // Invalidate all archives
        $archive_URLs = $this->get_all_archive_urls();
        if ( ! empty( $archive_URLs ) ) {
            $filter->addUrls( $archive_URLs );
        }

        $this->send_revalidation( $filter );
    }

    /**
     * @return string[]
     */
    private function get_all_archive_urls() {
        $archives_html = wp_get_archives( [ 'echo' => false, 'format' => 'link' ] );
        if ( ! preg_match_all( '#href=\'([^\']+)\'#', $archives_html, $matches ) ) {
            return [];
        }

        list( , $urls ) = $matches;

        return $urls;
    }

    /**
     * Sends an asset filter to the assets API.
     *
     * @param AssetFilter $filter
     */
    private function send_revalidation( AssetFilter $filter ) {
        // Do not revalidate if not logged in with Baqend
        if ( ! $this->plugin->baqend->isConnected() ) {
            return;
        }

        try {
            $this->plugin->baqend->asset()->revalidate( $filter );
            $this->logger->debug( 'Revalidation succeeded', [ 'filter' => $filter->jsonSerialize() ] );
        } catch ( \Exception $e ) {
            $this->logger->error( 'Revalidation failed with ' . get_class( $e ) . ': ' . $e->getMessage(), [ 'exception' => $e ] );
        }
    }
}
