<?php

namespace Baqend\WordPress;

use Baqend\SDK\Baqend;
use Baqend\SDK\Client\ClientInterface;
use Baqend\SDK\Exception\ResponseException;
use Monolog\Handler\ErrorLogHandler;
use Monolog\Logger;

/**
 * The main plugin class.
 *
 * @property string slug
 * @property string version
 * @property Loader loader
 * @property Logger logger
 * @property Options options
 * @property Baqend baqend
 * @property \Symfony\Component\Serializer\Serializer serializer
 * @property Service\AnalyzerService analyzer_service
 * @property Service\FormService form_service
 * @property Service\IOService io_service
 * @property Service\SpeedKitService speed_kit_service
 * @property Service\StatsService stats_service
 * @property Service\WooCommerceService woo_commerce_service
 * @property Service\UploadService upload_service
 * @property Controller\AdminController admin_controller
 * @property Controller\AjaxController ajax_controller
 * @property Controller\CronController cron_controller
 * @property Controller\FrontendController frontend_controller
 * @property Controller\TriggerController trigger_controller
 * @property Controller\DashboardController dashboard_controller
 * @property Env $env
 */
class Plugin {

    const CONTROLLER_TAG = 'baqend.controller';

    /**
     * @var string|null
     */
    public $app_name = null;

    /**
     * @var string
     */
    private $css_dir_url;

    /**
     * @var string
     */
    private $icons_dir_url;

    /**
     * @var string
     */
    private $js_dir_url;

    /**
     * @var string
     */
    private $css_dir_path;

    /**
     * @var string
     */
    private $js_dir_path;

    public function __construct() {
        $logger       = new Logger( 'baqend', [ new ErrorLogHandler() ] );
        $this->logger = $logger;

        $env       = new Env();
        $this->env = $env;

        $this->slug    = Info::SLUG;
        $this->version = Info::VERSION;

        $this->options    = new Options( $this );
        $this->loader     = new Loader( $this );
        $this->serializer = Baqend::createSerializer();
        $this->baqend     = new Baqend( ClientInterface::WORD_PRESS_TRANSPORT, [], null, $this->serializer );

        // Services
        $this->io_service           = new Service\IOService();
        $this->upload_service       = new Service\UploadService( $this->baqend, $this->io_service );
        $this->woo_commerce_service = new Service\WooCommerceService();
        $this->form_service         = new Service\FormService( $this );
        $this->analyzer_service     = new Service\AnalyzerService( $this->baqend->getClient(), $this->serializer, $this->options );
        $this->speed_kit_service    = new Service\SpeedKitService( $this->io_service, $this->baqend->getSpeedKit(), $this );
        $this->stats_service        = new Service\StatsService( $this );

        // Controllers
        $this->admin_controller = new Controller\AdminController( $this, $logger );
        $this->loader->add_controller( $this->admin_controller );
        $this->ajax_controller = new Controller\AjaxController( $this, $logger );
        $this->loader->add_controller( $this->ajax_controller );
        $this->frontend_controller = new Controller\FrontendController( $this, $logger );
        $this->loader->add_controller( $this->frontend_controller );
        $this->trigger_controller = new Controller\TriggerController( $this, $logger );
        $this->loader->add_controller( $this->trigger_controller );
        $this->cron_controller = new Controller\CronController( $this, $logger );
        $this->loader->add_controller( $this->cron_controller );
        $this->dashboard_controller = new Controller\DashboardController( $this, $logger );
        $this->loader->add_controller( $this->dashboard_controller );

        $this->css_dir_url   = plugin_dir_url( __DIR__ ) . 'css/';
        $this->js_dir_url    = plugin_dir_url( __DIR__ ) . 'js/';
        $this->css_dir_path  = plugin_dir_path( __DIR__ ) . 'css/';
        $this->js_dir_path   = plugin_dir_path( __DIR__ ) . 'js/';
        $this->icons_dir_url = plugin_dir_url( __DIR__ ) . 'icons/';

        // Load the text domain for i18n
        $this->loader->add_action( 'plugins_loaded', [ $this, 'load_textdomain' ] );

        // Register Baqend login and logout handlers
        $this->loader->add_action( 'plugins_loaded', [ $this, 'login' ] );
        $this->loader->add_action( 'plugins_loaded', [ $this, 'logout' ] );
        $this->loader->add_action( 'activated_plugin', [ $this, 'activated_plugin' ], 10, 1 );
        $this->loader->add_action( 'init', [ $this, 'handle_service_worker_request' ] );
    }

    /**
     * Builds a CSS URL for a given file.
     *
     * @param string $css_file
     *
     * @return string
     */
    public function css_url( $css_file ) {
        return $this->css_dir_url . $css_file;
    }

    /**
     * Builds an icon URL for a given file.
     *
     * @param string $icons_file
     *
     * @return string
     */
    public function icons_url( $icons_file ) {
        return $this->icons_dir_url . $icons_file;
    }

    /**
     * Builds a JavaScript URL for a given file.
     *
     * @param string $js_file
     *
     * @return string
     */
    public function js_url( $js_file ) {
        return $this->js_dir_url . $js_file;
    }

    /**
     * Builds a CSS path for a given file.
     *
     * @param string $css_file
     *
     * @return string
     */
    public function css_path( $css_file ) {
        return $this->css_dir_path . $css_file;
    }

    /**
     * Builds a JavaScript path for a given file.
     *
     * @param string $js_file
     *
     * @return string
     */
    public function js_path( $js_file ) {
        return $this->js_dir_path . $js_file;
    }

    /**
     * @return string
     */
    public function get_basename() {
        return plugin_basename( dirname( __DIR__ ) . '/baqend.php' );
    }

    /**
     * @return string
     */
    public function get_description() {
        return __( 'This plugin connects your WordPress website to Baqend and makes it super-fast.', 'baqend' );
    }

    /**
     * Runs the plugin.
     */
    public function run() {
        $this->loader->run();
    }

    /**
     * Sets the default options in the options table on activation.
     *
     * @throws \Exception
     */
    public function activate() {
        // Reset the options if they do not exist
        if ( $this->options->is_empty() ) {
            $this->options->reset();
        }

        // Checks that all necessary files exist
        $this->speed_kit_service->ensure_files_up_to_date();

        // Set hook to update Speed Kit
        $recurrence = 'hourly';
        $this->logger->debug( 'Set the update hook', [ 'recurrence' => $recurrence ] );
        wp_schedule_event( time(), $recurrence, 'cron_update_speed_kit' );
    }

    /**
     * Deactivates the plugin.
     */
    public function deactivate() {
        // Clear hook to update Speed Kit
        wp_clear_scheduled_hook( 'cron_update_speed_kit' );
    }

    /**
     * Executed once a plugin is activated.
     *
     * @param string $plugin The plugin which was activated.
     */
    public function activated_plugin( $plugin ) {
        if ( $plugin === $this->get_basename() ) {
            // Redirect to settings
            $location = $this->is_plesk_user() ? baqend_admin_url( 'baqend_speed_kit#activated' ) : baqend_admin_url( 'baqend#activated' );
            exit( wp_redirect( $location ) );
        }
    }

    /**
     * Handles the service worker request.
     */
    public function handle_service_worker_request() {
        $request_uri = $_SERVER['REQUEST_URI'];
        if ( $request_uri === extract_path_from_URL( $this->sw_url() ) ) {
            header( 'Content-Type: application/javascript' );
            header( 'Service-Worker-Allowed: ' . $this->sw_scope() );

            require $this->sw_path();

            exit;
        }
    }

    /**
     * Logs WordPress in to Baqend using a POST request or values from the settings.
     *
     * This method will be executed when the plugin is loaded.
     * It performs a request against Baqend to either login or validate the saved user token.
     */
    public function login() {
        $db       = $this->baqend;
        $app_name = null;

        if ( $this->options->has( 'app_name' ) && $this->options->has( 'authorization' ) ) {
            // Login from options setting
            $app_name      = $this->options->get( 'app_name' );
            $authorization = $this->options->get( 'authorization' );

            $db->connect( $app_name );
            $db->getRestClient()->setAuthorizationToken( $authorization );
        } elseif ( $this->options->has( 'app_name' ) && $this->options->has( 'api_token' ) ) {
            // Login from API token setting
            $app_name  = $this->options->get( 'app_name' );
            $api_token = $this->options->get( 'api_token' );

            $db->connect( $app_name );
            $db->getRestClient()->setAuthorizationToken( $api_token );
        }

        // Set invalid token handler
        if ( $db->isConnected() && $db->getRestClient()->isAuthorized() ) {
            $this->app_name = $app_name;
            if ( $this->options->has( 'api_token' ) ) {
                $db->getRestClient()->setInvalidTokenHandler( function () use ( $db ) {
                    // Set token again
                    $api_token = $this->options->get( 'api_token' );
                    $db->getRestClient()->setAuthorizationToken( $api_token );

                    // Update token in settings
                    $this->options->set( 'authorization', $db->getRestClient()->getAuthorizationToken() );
                    $this->options->save();
                } );
            }
        }
    }

    /**
     * Logs a user into his app.
     *
     * @param string $app_name The app name to login to.
     * @param string $username The username to use on that app.
     * @param string $password The user's password on that app.
     * @param string $bbq_username The username used in the Baqend Dashboard.
     * @param string $bbq_password The password used in the Baqend Dashboard.
     *
     * @return null|string The logged in app's name or null, if login failed.
     */
    public function login_user( $app_name, $username, $password, $bbq_username = '', $bbq_password = '' ) {
        $this->baqend->connect( $app_name );
        try {
            $this->baqend->user()->login( $username, $password );
        } catch ( ResponseException $e ) {
            return null;
        }

        $token = $this->baqend->getRestClient()->getAuthorizationToken();

        return $this->login_with_token( $app_name, $username, $token, $bbq_username, $bbq_password );
    }

    /**
     * Logs a user into his app using a token.
     *
     * @param string $app_name The app name to login to.
     * @param string $username The username to use on that app.
     * @param string $token The user's token on that app.
     * @param string $bbq_username The username used in the Baqend Dashboard.
     * @param string $bbq_password The password used in the Baqend Dashboard.
     *
     * @return null|string The logged in app's name or null, if login failed.
     */
    public function login_with_token( $app_name, $username, $token, $bbq_username = '', $bbq_password = '' ) {
        // Save app to options
        if ( $token ) {
            $this->options->set( 'app_name', $app_name );
            $this->options->set( 'username', $username );

            if ( $bbq_username ) {
                $this->options->set( 'bbq_username', $bbq_username );
            } else {
                $this->options->remove( 'bbq_username' );
            }
            if ( $bbq_password ) {
                $this->options->set( 'bbq_password', $bbq_password );
            } else {
                $this->options->remove( 'bbq_password' );
            }

            $this->options->set( 'destination_scheme', 'https' );
            $this->options->set( 'destination_host', $app_name . '.app.baqend.com' );
            $this->options->set( 'authorization', $token );
            $this->options->set( 'speed_kit_enabled', true );
            $this->options->save();

            return $app_name;
        }

        return null;
    }

    /**
     * Logs WordPress out from Baqend using a POST request.
     *
     * This method will be executed when the plugin is loaded.
     */
    public function logout() {
        $db = $this->baqend;
        if ( $db->isConnected() && array_key_exists( 'logout', $_POST ) && $_POST['logout'] === 'true' ) {
            // Execute logout request
            $this->did_logout();
            $this->app_name = null;
            $this->options->remove( 'app_name' );
            $this->options->remove( 'authorization' );
            $this->options->remove( 'bbq_password' );
            $this->options->remove( 'bbq_username' );
            $this->options->save();
        }
    }

    /**
     * Loads the plugin language files.
     */
    public function load_textdomain() {
        load_plugin_textdomain(
            Info::SLUG,
            false,
            dirname( dirname( plugin_basename( __FILE__ ) ) ) . '/languages/'
        );
    }

    /**
     * Returns the Service Worker's absolute pathname.
     *
     * @return string
     */
    public function sw_path() {
        return trailingslashit( ABSPATH ) . 'speed-kit-sw.js';
    }

    /**
     * Returns the Service Worker's absolute URL.
     *
     * @return string
     */
    public function sw_url() {
        return home_url( 'speed-kit-sw.js' );
    }

    /**
     * Returns the service worker's scope.
     *
     * @return string
     */
    public function sw_scope() {
        $home = home_url( '/', 'https' );
        $site = site_url( '/', 'https' );

        // Return home url when the both host are not the same.
        if ( parse_url( $home )['host'] !== parse_url( $site )['host'] ) {
            return $home;
        }

        for ( $i = 0; $i < min( strlen( $home ), strlen( $site ) ); $i += 1 ) {
            if ( $home[ $i ] !== $site[ $i ] ) {
                break;
            }
        }

        return substr( $home, 0, $i );
    }

    /**
     * Returns the snippet's absolute pathname.
     *
     * @return string
     */
    public function snippet_path() {
        return $this->js_path( 'snippet.js' );
    }

    /**
     * Returns the Service Worker's absolute pathname in the JS dir.
     *
     * @return string
     */
    public function sw_js_path() {
        return $this->js_path( 'sw.js' );
    }

    /**
     * Returns the dynamic fetcher's absolute pathname.
     *
     * @return string
     */
    public function dynamic_fetcher_path() {
        return $this->js_path( 'dynamic-fetcher.js' );
    }

    /**
     * @return bool
     */
    public function is_plesk_user() {
        return $this->options->has( 'app_name' ) && $this->options->has( 'api_token' );
    }

    /**
     * Trigger a cache update in WordPress.
     */
    public function trigger_wp_cache_invalidation() {
        // Hack: Let other caching plugins revalidate the generated HTML
        switch_theme( wp_get_theme()->get_stylesheet() );
    }

    /**
     * @return bool
     */
    private function did_logout() {
        try {
            return $this->baqend->user()->logout();
        } catch ( \Exception $e ) {
            return true;
        }
    }

    /**
     * Returns a POST value.
     *
     * @param string $field
     * @param string $default
     *
     * @return string
     */
    private function get_post( $field, $default = '' ) {
        if ( isset( $_POST[ $field ] ) ) {
            return $_POST[ $field ];
        }

        return $default;
    }
}
